#ifndef _EXPR_H_
#define _EXPR_H_

#include "stdafx.h"

#define NUMBER                0
#define ARGUMENT              1
#define OPERATOR              2
#define FUNCTION              3

#define ADDITION              '+'
#define SUBTRACTION           '-'
#define MULTIPLICATION        '*'
#define DIVISION              '/'
#define POWER                 '^'
#define LEFT_BRACKET          '('
#define RIGHT_BRACKET         ')'

#define BRACKETS              0
#define SQRT                  1
#define SIN                   2
#define COS                   3
#define TAN                   4
#define CTAN                  5
#define LOG                   6
#define LN                    7
#define EXP                   8
#define ABS                   9
#define SIGN                  10

#define PRIOR_NUMBER          0
#define PRIOR_ARGUMENT        0
#define PRIOR_ADDITION        1
#define PRIOR_SUBTRACTION     1
#define PRIOR_MULTIPLICATION  2
#define PRIOR_DIVISION        2
#define PRIOR_POWER           3
#define PRIOR_FUNCTION        4
#define PRIOR_BRACKETS        0

#define MAX_TOKEN             100

struct Token {
  Token() { }
  int type;
  int type2;
  int priority;
  double value;
};

class Expr {
    Token list[MAX_TOKEN];
    int end, p1, p2;
  public:
    Expr();
    void AddToken(Token);
    Token GetToken();
    Expr GetOpArg(int);
    Expr GetFuncArg();
    int End() { return end; }
};

Expr::Expr()
{
  end = true;
  p1 = p2 = 0;
}

void Expr::AddToken(Token t)
{
  list[p2++] = t;
  end = false;
}

Token Expr::GetToken()
{
  if (p2 == p1 + 1) end = true;
  return list[p1++];
}

Expr Expr::GetOpArg(int priority)
{
  Expr e;
  Token t = list[p1];

  while (t.priority > priority || !t.priority)
  {
    e.AddToken(t);
    GetToken();
    if (t.type == FUNCTION)
    {
      t = GetToken();
      while (t.type2 != RIGHT_BRACKET)
      {
        e.AddToken(t);
        t = GetToken();
      }
      t.type = OPERATOR;
      t.type2 = RIGHT_BRACKET;
      t.priority = PRIOR_BRACKETS;
      e.AddToken(t);
    }
    if (end) break;
    t = list[p1];
  }

  return e;
}

Expr Expr::GetFuncArg()
{
  Expr e;
  Token t;

  t = GetToken();
  while (t.type2 != RIGHT_BRACKET)
  {
    e.AddToken(t);
    if (t.type == FUNCTION)
    {
      Expr new_expr = GetFuncArg();
      do
      {
        t = new_expr.GetToken();
        e.AddToken(t);
      } while (!new_expr.End());
      t.type = OPERATOR;
      t.type2 = RIGHT_BRACKET;
      t.priority = PRIOR_BRACKETS;
      e.AddToken(t);
    }
    if (end) break;
    t = GetToken();
  }

  return e;
}

class ExprException {
    char *s;
  public:
    ExprException() { s = NULL; }
    ExprException(char *c)
    {
      s = new char[strlen(c) + 1];
      strcpy(s, c);
    }
    ~ExprException()
    {
      delete s;
    }
    virtual void GetMessage(char *msg)
    {
      strcpy(msg, s);
    }
};

class UnknownSymbol: public ExprException {
    int pos;
    char s;
  public:
    UnknownSymbol(int p, char c)
    {
      pos = p;
      s = c;
    }
    void GetMessage(char *msg)
    {
      sprintf(msg, "%d:   '%c'", pos + 1, s);
    }
};

class UnknownFunction: public ExprException {
    int pos;
    char *s;
  public:
    UnknownFunction(int p, char *c)
    {
      pos = p;
      s = new char[strlen(c) + 1];
      strcpy(s, c);
    }
    ~UnknownFunction()
    {
      delete s;
    }
    void GetMessage(char *msg)
    {
      sprintf(msg, "%d:   \"%s\"", pos + 1, s);
    }
};

class DivideError: public ExprException {
  public:
    DivideError() { }
    void GetMessage(char *msg)
    {
      strcpy(msg, "  0");
    }
};

class SqrtError: public ExprException {
    double v;
  public:
    SqrtError(double d) { v = d; }
    void GetMessage(char *msg)
    {
      sprintf(msg, " '%f'  ", (float)v);
    }
};

class LogError: public ExprException {
    double v;
  public:
    LogError(double d) { v = d; }
    void GetMessage(char *msg)
    {
      sprintf(msg, " '%f'     0", (float)v);
    }
};

class TanError: public ExprException {
    double v;
  public:
    TanError(double d) { v = d; }
    void GetMessage(char *msg)
    {
      sprintf(msg, "  (arg = %f)", (float)v);
    }
};

class CTanError: public ExprException {
    double v;
  public:
    CTanError(double d) { v = d; }
    void GetMessage(char *msg)
    {
      sprintf(msg, "  (arg = %f)", (float)v);
    }
};

class StringExpr {
    Expr main_expr;
    double GetValue(Expr, double);
    double Operator(Token, double, double);
    double Function(Token, double);
    double arg;
  public:
    StringExpr(char *);
    double GetValue(double arg)
    {
      StringExpr::arg = arg;
      return GetValue(main_expr, 0);
    }
};

StringExpr::StringExpr(char *s)
{
  Expr e;
  Token t;
  char *p = s;
  int prev_type = 0;

  for (unsigned i = 0; i < strlen(s); i++)
    if (isspace(s[i]))
    {
      for (unsigned j = i; j < strlen(s); j++)
        s[j] = s[j + 1];
      i--;
    }

  while (*s)
  {
    while (isspace(*s)) s++;
    if (!*s) continue;
    if (strchr("+-*/^()", *s))
    {
      t.type = OPERATOR;
      t.type2 = *s;
      switch (*s)
      {
        case '+': t.priority = PRIOR_ADDITION; break;
        case '-': t.priority = PRIOR_SUBTRACTION; break;
        case '*': t.priority = PRIOR_MULTIPLICATION; break;
        case '/': t.priority = PRIOR_DIVISION; break;
        case '^': t.priority = PRIOR_POWER; break;
        case ')': t.priority = PRIOR_BRACKETS;
      }
      s++;
    }
    else if (toupper(*s) == 'X')
    {
      t.type = ARGUMENT;
      t.type2 = t.type;
      t.priority = PRIOR_ARGUMENT;
      s++;
    }
    else if (isdigit(*s))
    {
      int p = 0;
      char temp[20], *r = temp;
      while (isdigit(*s) || (*s == '.' && !p))
      {
        if (*s == '.') p = 1;
        *r++ = *s++;
      }
      *r = 0;
      t.type = NUMBER;
      t.type2 = t.type;
      t.priority = PRIOR_NUMBER;
      t.value = atof(temp);
    }
    else if (isalpha(*s))
    {
      char temp[20], *r = temp;
      while (isalpha(*s) || isdigit(*s)) *r++ = *s++;
      *r = 0;
      t.type = FUNCTION;
      t.priority = PRIOR_FUNCTION;
      for (unsigned i = 0; i < strlen(temp); i++)
        temp[i] = toupper(temp[i]);
      if (!strcmp(temp, "SQRT"))
        t.type2 = SQRT;
      else if (!strcmp(temp, "SIN"))
        t.type2 = SIN;
      else if (!strcmp(temp, "COS"))
        t.type2 = COS;
      else if (!strcmp(temp, "TAN"))
        t.type2 = TAN;
      else if (!strcmp(temp, "CTAN"))
        t.type2 = CTAN;
      else if (!strcmp(temp, "LOG"))
        t.type2 = LOG;
      else if (!strcmp(temp, "LN"))
        t.type2 = LN;
      else if (!strcmp(temp, "EXP"))
        t.type2 = EXP;
      else if (!strcmp(temp, "ABS"))
        t.type2 = ABS;
      else if (!strcmp(temp, "SIGN"))
        t.type2 = SIGN;
      else if (!strcmp(temp, "PI"))
      {
        t.type = NUMBER;
        t.type2 = t.type;
        t.priority = PRIOR_NUMBER;
        t.value = M_PI;
      }
      else if (!strcmp(temp, "E"))
      {
        t.type = NUMBER;
        t.type2 = t.type;
        t.priority = PRIOR_NUMBER;
        t.value = M_E;
      }
      else
        throw new UnknownFunction(s - p - strlen(temp), temp);
    }
    else
      throw new UnknownSymbol(s - p - 1, *s);
    e.AddToken(t);
  }

  prev_type = 0;

  do
  {
    t = e.GetToken();
    if (t.type2 == LEFT_BRACKET)
    {
      if (prev_type != FUNCTION)
      {
        t.type = FUNCTION;
        t.type2 = BRACKETS;
        t.priority = PRIOR_FUNCTION;
        main_expr.AddToken(t);
      }
      prev_type = 0;
    }
    else
    {
      main_expr.AddToken(t);
      prev_type = t.type;
    }
  } while (!e.End());
}

double StringExpr::GetValue(Expr e, double prev_res)
{
  Token t;
  double res = prev_res;

  while (!e.End())
  {
    t = e.GetToken();
    switch (t.type)
    {
      case NUMBER: res = t.value; break;
      case ARGUMENT: res = arg; break;
      case OPERATOR: res = Operator(t, res, GetValue(e.GetOpArg(t.priority), 0)); break;
      case FUNCTION: res = Function(t, GetValue(e.GetFuncArg(), 0));
    }
  }

  return res;
}

double StringExpr::Operator(Token t, double v1, double v2)
{
  switch (t.type2)
  {
    case ADDITION: return v1 + v2;
    case SUBTRACTION: return v1 - v2;
    case MULTIPLICATION: return v1 * v2;
    case DIVISION: if (!v2 && v1) throw new DivideError(); else return (!v1 && !v2) ? 1 : v1 / v2;
    case POWER: return pow(v1, v2);
  }

  return 0;
}

double StringExpr::Function(Token t, double arg)
{
  switch (t.type2)
  {
    case BRACKETS: return arg;
    case SQRT: if (arg < 0) throw new SqrtError(arg); else return sqrt(arg);
    case SIN: return sin(arg);
    case COS: return cos(arg);
    case TAN: if (arg == M_PI / 2 || arg == -M_PI / 2) throw new TanError(arg); else return tan(arg);
    case CTAN: if (!arg || arg == M_PI) throw new CTanError(arg); else return 1 / tan(arg);
    case LOG: if (arg <= 0) throw new LogError(arg); else return log10(arg);
    case LN: if (arg <= 0) throw new LogError(arg); else return log(arg);
    case EXP: return exp(arg);
    case ABS: return fabs(arg);
    case SIGN: if (!arg) return 0; else return arg < 0 ? -1 : 1;
  }

  return 0;
}

#endif

